PROGRAM LISP(INPUT,OUTPUT);

(* written by W. Taylor and L. Cox *)

CONST MAXNODE=2000;
  DEFINEDWORDS=22;

TYPE 
  INPUTSYMBOL = 
    (ATOM, PERIOD, LPAREN, RPAREN);
  RESERVDWORDS =
    (LABELSYM, FUNARSYM, LAMBDSYM, QUOTESYM, CONDSYM, FUNCTSYM,
     DEFINSYM, SETQSYM ,
     ATOMSYM, NUMBPSYM, CARSYM, CDRSYM, COPYSYM,
     CONSSYM, EQSYM, REPLASYM, REPLDSYM, APPENSYM, ADDSYM,
     SUBSYM, TIMESYM, DIVSYM, EVALSYM, APPLYSYM,
     LASTSYM);
  STATUSTYPE = 
    (UNMARKED, LEFT, RIGHT, MARKED);
  NODEKIND= (SEXPR, ANATOM, NUMERIC);
  PROPKIND= (PNAME, PEXPR, PFEXPR, PSUBR, PFSUBR, PVALUE, PNONE);
  PROPSET = SET OF PROPKIND;
  SYMBEXPPTR = ^SYMBOLICEXPRESSION;
  PROPPTR   = ^PROPERTYNODE;
  ALFA = ARRAY[1..10] OF CHAR;
  SYMBOLICEXPRESSION = RECORD
     STATUS : STATUSTYPE;
     NEXT   : SYMBEXPPTR;
     CASE KIND:NODEKIND OF
       ANATOM: (PROPLIST: PROPPTR; RESSYM: RESERVDWORDS);
       NUMERIC:(NVALUE  : INTEGER);
       SEXPR : (HEAD, TAIL: SYMBEXPPTR)
     END;
  PROPERTYNODE = RECORD
     NEXT : PROPPTR;
     CASE KIND: PROPKIND OF
       PEXPR, PFEXPR: (PNODE: SYMBEXPPTR);
      PSUBR, PFSUBR:
       (RESSYM: RESERVDWORDS; ARGS: INTEGER);
      PNAME: ( NAME: ALFA)
     END;

 (*       T H E   G L O B A L   V A R I A B L E S  *)

VAR 

 (* variables that pass information from the scanner to the 
    read routine *)

  LOOKAHEADSYM, SYM : INPUTSYMBOL;
  ID: ALFA;
  ALREADYPEEKED: BOOLEAN;
  CH: CHAR;
  PTR: SYMBEXPPTR; (* EXPRESSION BEING EVALUATED *)

 (* the global lists of LISP nodes *)

  FREELIST, NODELIST: SYMBEXPPTR;
  PPTR, NOPROP, OBPROP: PROPPTR;

 (* The list of names *)

  OBLIST : SYMBEXPPTR;

 (* the nodes with constant values *)

  FINATOM, NILATOM, TATOM, FNARGATOM: SYMBEXPPTR;

 (* variables used to identify atoms with predefined meanings *)

  RESWORD: RESERVDWORDS;
  RESWORDS: ARRAY[RESERVDWORDS] OF ALFA;
  FREENODES: INTEGER;
  NUMBEROFGCS: INTEGER; (* number of garbage collections made *)

  EVALVAL: SYMBEXPPTR;
  ERRORS : BOOLEAN;


VALUE
  RESWORDS = (
    'LABEL     ', 'FUNARG    ', 'LAMBDA    ',
    'QUOTE     ', 'COND      ', 'FUNCTION  ',
    'DEFINE    ', 'SETQ      ',
    'ATOM      ', 'NUMBERP   ', 'CAR       ',
    'CDR       ', 'COPY      ', 'CONS      ',
    'EQ        ', 'REPLACA   ', 'REPLACD   ',
    'APPEND    ', 'ADD       ', 'SUB       ',
    'TIMES     ', 'DIV       ', 'EVAL      ',
    'APPLY     ',
    '          ');



  PROCEDURE GARBAGEMAN;


   PROCEDURE MARK(LIST: SYMBEXPPTR);
    VAR FATHER, SON, CURRENT: SYMBEXPPTR;
      CHEAD, CTAIL: SYMBEXPPTR;

   BEGIN (* MARK *)
   FATHER:=NIL; CURRENT:=LIST; SON:=CURRENT;
   WHILE CURRENT<>NIL DO
    BEGIN CHEAD:=CURRENT^.HEAD; CTAIL:=CURRENT^.TAIL;
     CASE CURRENT^.STATUS OF
       UNMARKED:
         IF CURRENT^.KIND>SEXPR THEN CURRENT^.STATUS:=MARKED
         ELSE
          IF (CURRENT^.HEAD^.STATUS<>UNMARKED) OR
             (CHEAD=CURRENT) THEN
            IF (CTAIL^.STATUS<>UNMARKED) OR
               (CTAIL=CURRENT)
            THEN CURRENT^.STATUS:=MARKED
            ELSE BEGIN CURRENT^.STATUS:=RIGHT; SON:=CTAIL;
              CURRENT^.TAIL:=FATHER; FATHER:=CURRENT; CURRENT:=SON
              END
          ELSE BEGIN CURRENT^.STATUS:=LEFT; SON:=CHEAD;
            CURRENT^.HEAD:=FATHER; FATHER:=CURRENT; CURRENT:=SON
            END;
       LEFT:
         IF CTAIL^.STATUS<>UNMARKED THEN
           BEGIN CURRENT^.STATUS:=MARKED; FATHER:=CHEAD;
           CURRENT^.HEAD:=SON; SON:=CURRENT
           END
         ELSE BEGIN CURRENT^.STATUS:=RIGHT; CURRENT^.HEAD:=SON;
           CURRENT^.TAIL:=CHEAD; CURRENT:=CTAIL; SON:=CURRENT
           END;
       RIGHT:
         BEGIN CURRENT^.STATUS:=MARKED; FATHER:=CTAIL;
         CURRENT^.TAIL:=SON; SON:=CURRENT
         END;
       MARKED: CURRENT:=FATHER
     END (* CASE *)
    END
   END (* MARK *);

   PROCEDURE COLLECTFREENODES;

   VAR TEMP: SYMBEXPPTR;

   BEGIN (* COLLECTFREENODES *)
   WRITELN(' number of free nodes before collection = ',
           FREENODES:1, '.');
   FREELIST:=NIL;  FREENODES:=0;  TEMP:=NODELIST;
   WHILE TEMP<>NIL DO
     BEGIN 
     IF (TEMP^.STATUS <> UNMARKED) OR 
        (TEMP^.KIND   =  ANATOM  ) THEN TEMP^.STATUS:=UNMARKED
     ELSE BEGIN FREENODES:=FREENODES+1;
       TEMP^.HEAD:=FREELIST; FREELIST:=TEMP
       END;
     TEMP:=TEMP^.NEXT
     END;
   WRITELN(' number of freenodes after collection = ',
           FREENODES:1, '.')
   END (* COLLECTFREENODES *);


  BEGIN (* GARBAGEMAN *)
  NUMBEROFGCS:=NUMBEROFGCS+1; WRITELN;
  WRITELN(' garbage collection. '); WRITELN; MARK(OBLIST);
  IF PTR<>NIL THEN MARK(PTR); COLLECTFREENODES
  END (* GARBAGEMAN *);


  PROCEDURE RETURN(LEVEL: INTEGER); EXTERN;


  PROCEDURE POP(VAR SPTR: SYMBEXPPTR);

  BEGIN 
    IF FREELIST = NIL THEN
      BEGIN
        WRITELN(' Heap overflow!! ');
        ERRORS:=TRUE; RETURN(1)
      END;
    FREENODES:=FREENODES-1; SPTR:=FREELIST;
    FREELIST := FREELIST^.HEAD;
    SPTR^.KIND:=SEXPR;
    SPTR^.HEAD:=NIL; SPTR^.TAIL:=NIL
  END (* POP *);


 (* input output utility routines *)


  PROCEDURE ERROR(NUMBER: INTEGER);

  BEGIN 
    WRITELN;  WRITE('  Error  ',NUMBER:1, ',');
    ERRORS:=TRUE;
    CASE NUMBER OF
     1: WRITELN(' Atom or Lparen expected in the S-expr.');
     2: WRITELN(' Atom, Lparen, or Rparen expected in the S-expr.');
     3: WRITELN(' Illegal special function.');
     4: WRITELN(' Rparen expected in the S-expr.');
     5: WRITELN(' 1st argument of REPLACA is an atom.');
     6: WRITELN(' 1st argument of REPLACD is an atom.');
     7: WRITELN(' Argument of CAR is an atom.');
     8: WRITELN(' Argument of CDR is an atom.');
     9: WRITELN(' Argument of APPEND is not list.');
    10: WRITELN(' Comma or Rparen expected in concatenate.');
    11: WRITELN(' EOF before FIN.');
    12: WRITELN(' Lambda or Label expected.');
    13: WRITELN(' Error in converting number.');
    14: WRITELN(' Expecting function.');
    15: WRITELN(' Expecting VAR.')
   END;
   IF NUMBER = 11 THEN HALT(' LABEL 2');
   RETURN(1)
   END (* ERROR *);


   FUNCTION GETPROP(ATOM: SYMBEXPPTR; PROPS: PROPSET): PROPPTR;
   VAR P: PROPPTR;

   BEGIN
     GETPROP:=NOPROP;
     P:=ATOM^.PROPLIST;
     IF ATOM^.KIND=ANATOM THEN
       WHILE (P<>NIL) DO
         IF P^.KIND IN PROPS THEN
           BEGIN GETPROP:=P; P:=NIL END
         ELSE P:=P^.NEXT;
    END;


    PROCEDURE SETPROP(VAR P: PROPPTR;
       ATOM: SYMBEXPPTR; PROP: PROPKIND);

    BEGIN P:=GETPROP(ATOM,[PROP]);
      IF P=NOPROP THEN
        BEGIN NEW(P); P^.NEXT:=ATOM^.PROPLIST;
          ATOM^.PROPLIST:=P; P^.KIND:=PROP
        END
    END;


  PROCEDURE BACKUPINPUT;
    BEGIN ALREADYPEEKED:=TRUE; LOOKAHEADSYM:=SYM; SYM:=LPAREN
    END (* BACKUPINPUT *);




  FUNCTION FORMATOM(N: ALFA): SYMBEXPPTR;

  VAR ATOM: SYMBEXPPTR; PPTR: PROPPTR;


    PROCEDURE FINDATOM(VAR PTR: SYMBEXPPTR);

    BEGIN
      IF PTR=NILATOM THEN
       BEGIN POP(PTR); POP(ATOM); PTR^.TAIL:=ATOM;
         ATOM^.TAIL:=NILATOM; ATOM^.HEAD:=NILATOM;
         POP(ATOM); PTR^.HEAD:=ATOM; ATOM^.KIND:=ANATOM;
         ATOM^.PROPLIST:=NIL; SETPROP(PPTR,ATOM,PNAME);
         PPTR^.NAME:=N
       END
      ELSE BEGIN PPTR:=GETPROP(PTR^.HEAD,[PNAME]);
         IF PPTR^.NAME=N THEN ATOM:=PTR^.HEAD
         ELSE
          IF PPTR^.NAME<N 
          THEN FINDATOM(PTR^.TAIL^.HEAD)
          ELSE FINDATOM(PTR^.TAIL^.TAIL)
         END
     END (* FINDATOM *);

     PROCEDURE CONVERTNUMBER(VAR V: INTEGER);
     VAR I, J: INTEGER;

     BEGIN
     V:=0;
     FOR I:=1 TO 10 DO 
      IF N[I] IN ['0'..'9'] THEN 
         V:=V*10+ORD(N[I])-ORD('0')
      ELSE IF N[I]<>' ' THEN ERROR(13);
     END (* CONVERT NUMBER *);


  BEGIN
  IF N[1] IN ['0'..'9'] THEN
    BEGIN POP(ATOM); ATOM^.KIND:=NUMERIC;
      CONVERTNUMBER(ATOM^.NVALUE)
    END
  ELSE FINDATOM(OBLIST);
  FORMATOM:=ATOM
  END (* FORMATOM *);



  PROCEDURE NEXTSYM;

  VAR I: INTEGER;

  BEGIN 
  IF ALREADYPEEKED
  THEN BEGIN SYM:=LOOKAHEADSYM; ALREADYPEEKED:=FALSE END
  ELSE 
    BEGIN 
    WHILE CH = ' ' DO
      READ(CH);
    IF CH IN ['(','.',')']
    THEN BEGIN
      CASE CH OF
       '(': SYM:=LPAREN;
       '.': SYM:=PERIOD;
       ')': SYM:=RPAREN;
      END (* CASE *);
      READ(CH);
      END
    ELSE BEGIN SYM:=ATOM; ID:='          '; I:=0;
      REPEAT I:=I+1; 
        IF I<11 THEN ID[I]:=CH;
        READ(CH)
      UNTIL CH IN [' ','(','.',')']
      END;
    END;
  END (* NEXTSYM *);




  PROCEDURE READEXPR(VAR SPTR: SYMBEXPPR);

  VAR NXT: SYMBEXPPTR;

  BEGIN 
  POP(SPTR);
  CASE SYM OF 
    RPAREN, PERIOD: ERROR(1);
    ATOM: SPTR:=FORMATOM(ID);
    LPAREN:
     BEGIN NEXTSYM;
       IF SYM=PERIOD THEN ERROR(2)
       ELSE 
        IF SYM=RPAREN 
        THEN BEGIN SPTR^.HEAD:=FREELIST;
            FREELIST:=SPTR;  SPTR:=NILATOM;
          END
        ELSE 
         BEGIN READEXPR(SPTR^.HEAD); NEXTSYM;
           IF SYM=PERIOD
           THEN BEGIN NEXTSYM; READEXPR(SPTR^.TAIL); NEXTSYM;
             IF SYM<>RPAREN THEN ERROR(4)
             END
           ELSE BEGIN BACKUPINPUT; READEXPR(SPTR^.TAIL) END
         END
      END
     END (* CASE *);
   END (* READEXPR *);




  PROCEDURE PRINTNAME(ATOM: SYMBEXPPTR);
  VAR I:INTEGER;
    NAME: ALFA;
    P: PROPPTR;

  BEGIN
    P:=GETPROP(ATOM,[PNAME]);
    NAME:=P^.NAME;
    I:=1;
    REPEAT WRITE(NAME[I]); I:=I+1 UNTIL (NAME[I]=' ') OR (I=11);
    WRITE(' ')
  END (* PRINTNAME *);



  PROCEDURE PRINTEXPR(SPTR: SYMBEXPPTR);

  VAR FLAG: BOOLEAN;
    CTAIL: SYMBEXPPTR;

  BEGIN 
    CASE SPTR^.KIND OF
      ANATOM: PRINTNAME(SPTR);
      NUMERIC: WRITE(SPTR^.NVALUE:1,' ');
      SEXPR  :
       BEGIN WRITE('(');
       FLAG:=TRUE;
       WHILE FLAG DO
         BEGIN PRINTEXPR(SPTR^.HEAD);
         FLAG:=FALSE;
         CTAIL:=SPTR^.TAIL;
         CASE CTAIL^.KIND OF 
           NUMERIC: WRITE('.',CTAIL^.NVALUE:1,' )');
           ANATOM :
             IF CTAIL=NILATOM THEN WRITE(')')
             ELSE BEGIN WRITE('.'); PRINTEXPR(CTAIL);
               WRITE(')');
               END;
           SEXPR  : BEGIN SPTR:=CTAIL; FLAG:=TRUE END;
         OTHERWISE WRITE('???')
         END END (* WHILE *);
     END;
     OTHERWISE WRITE('???')
     END (* CASE *)
  END (* PRINTEXPR *);


 (*    T h e  E x p r e s s i o n  E v a l u a t e r  E v a l *)


  FUNCTION EVAL(E, ALIST: SYMBEXPPTR): SYMBEXPPTR;

  VAR
    TEMP, CAROFE, CAAROFE: SYMBEXPPTR;
    RESVAL : RESERVDWORDS;


    FUNCTION REPLACH(SPTR1, SPTR2: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN 
      IF SPTR1^.KIND=SEXPR THEN SPTR1^.HEAD:=SPTR2 ELSE ERROR(5);
      REPLACH:=SPTR1
    END (* REPLACH *);


    FUNCTION REPLACT(SPTR1, SPTR2: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN 
      IF SPTR1^.KIND=SEXPR THEN SPTR1^.TAIL:=SPTR2 ELSE ERROR(6);
      REPLACT:=SPTR1
    END (* REPLACET *);


    FUNCTION HEAD(SPTR: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN
      IF SPTR^.KIND=SEXPR THEN HEAD:=SPTR^.HEAD ELSE ERROR(7);
    END (* HEAD *);


    FUNCTION TAIL(SPTR: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN
      IF SPTR^.KIND=SEXPR THEN TAIL:=SPTR^.TAIL ELSE ERROR(8);
    END (* TAIL *);


    FUNCTION CONS(SPTR1, SPTR2: SYMBEXPPTR): SYMBEXPPTR;

    VAR TEMP: SYMBEXPPTR;

    BEGIN POP(TEMP); TEMP^.HEAD:=SPTR1;
      TEMP^.TAIL:=SPTR2; CONS:=TEMP
    END (* CONS *);


    FUNCTION COPY(SPTR: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN
      IF SPTR^.KIND=SEXPR THEN
         COPY:=CONS(COPY(SPTR^.HEAD),COPY(SPTR^.TAIL))
      ELSE COPY:=SPTR;
    END (* COPY *);


    FUNCTION APPEND(SPTR1, SPTR2: SYMBEXPPTR): SYMBEXPPTR;

    BEGIN 
      IF SPTR1^.KIND = SEXPR THEN
        APPEND:=CONS(COPY(SPTR1^.HEAD),APPEND(SPTR1^.TAIL, SPTR2))
      ELSE 
        IF SPTR1=NILATOM THEN APPEND:=SPTR2
           ELSE ERROR(9);
    END (* APPEND *);


    FUNCTION PREDICATE(R: BOOLEAN): SYMBEXPPTR;

    BEGIN
      IF R THEN PREDICATE :=TATOM ELSE PREDICATE :=NILATOM
    END;


    FUNCTION EQQ(SPTR1, SPTR2: SYMBEXPPTR): SYMBEXPPTR;
    VAR TRUTH: BOOLEAN;

    BEGIN
      TRUTH:=SPTR1^.KIND=SPTR1^.KIND;
      IF TRUTH THEN
        IF SPTR1^.KIND = NUMERIC
          THEN TRUTH:=SPTR1^.NVALUE=SPTR2^.NVALUE
          ELSE TRUTH:=SPTR1=SPTR2;
      EQQ:=PREDICATE(TRUTH);
    END (* EQQ *);


    FUNCTION LOOKUP(KEY, ALIST: SYMBEXPPTR; ERR: BOOLEAN)
        : SYMBEXPPTR;

    VAR FOUND: BOOLEAN;

    BEGIN
     FOUND:=FALSE;
      WHILE NOT FOUND AND (ALIST<>NILATOM) DO
       BEGIN
        FOUND:=ALIST^.HEAD^.HEAD=KEY;
        LOOKUP:=ALIST^.HEAD;
        ALIST:=ALIST^.TAIL;
       END;
      IF NOT FOUND THEN
       IF ERR THEN ERROR(15) ELSE LOOKUP:=NILATOM
    END (* LOOKUP *);


   FUNCTION EVLIS(M :SYMBEXPPTR): SYMBEXPPTR;

   BEGIN 
    IF M^.KIND=SEXPR THEN 
      EVLIS:=CONS(EVAL(M^.HEAD,ALIST),EVLIS(M^.TAIL))
    ELSE EVLIS:=M
   END;


    FUNCTION EVCON(CONDPAIRS: SYMBEXPPTR): SYMBEXPPTR;

    VAR TEMP, ONEPAIR: SYMBEXPPTR;

    BEGIN 
      REPEAT ONEPAIR:=HEAD(CONDPAIRS);
        TEMP:=EVAL(HEAD(ONEPAIR),ALIST);
        CONDPAIRS:=TAIL(CONDPAIRS)
      UNTIL TEMP<>NILATOM;
      EVCON:=EVAL(HEAD(TAIL(ONEPAIR)),ALIST)
    END (* EVCON *);


  FUNCTION APPLY(FN, ARGS, ALIST: SYMBEXPPTR): SYMBEXPPTR;
  VAR CADR, CADDR: SYMBEXPPTR;
      AGAIN: BOOLEAN;


    FUNCTION MATH(V: INTEGER): SYMBEXPPTR;
    VAR TEMP: SYMBEXPPTR;

    BEGIN
      IF (CADR^.KIND=NUMERIC) AND (CADDR^.KIND=NUMERIC)
      THEN BEGIN POP(TEMP);
        TEMP^.KIND:=NUMERIC; TEMP^.NVALUE:=V
        END
      ELSE TEMP:=NILATOM;
    MATH:=TEMP
    END;


    FUNCTION BINDARGS(NAMES,VALUES: SYMBEXPPTR): SYMBEXPPTR;

    VAR TEMP, TEMP2: SYMBEXPPTR;

    BEGIN 
      TEMP:=ALIST;
      WHILE NAMES<>NILATOM DO
        BEGIN TEMP:=CONS(VALUES,TEMP);
          VALUES:=VALUES^.TAIL;
          TEMP^.HEAD^.TAIL:=TEMP^.HEAD^.HEAD;
          TEMP^.HEAD^.HEAD:=NAMES^.HEAD;
          NAMES:=NAMES^.TAIL;
        END;
      BINDARGS:=TEMP
    END (* BINDARGS *);


  BEGIN APPLY:=NILATOM;
    AGAIN:=TRUE;
    WHILE AGAIN DO
    IF FN^.KIND = ANATOM THEN 
     BEGIN PPTR:=GETPROP(FN,[PEXPR,PSUBR]);
       CASE PPTR^.KIND OF
         PEXPR: FN:=PPTR^.PNODE;
         PNONE: FN:=TAIL(LOOKUP(FN,ALIST,TRUE));
         PSUBR:
          BEGIN CADR:=HEAD(ARGS);
          AGAIN:=FALSE;
          IF PPTR^.ARGS=1 THEN CADDR:=HEAD(TAIL(ARGS));
          CASE PPTR^.RESSYM OF
            ATOMSYM: APPLY:=PREDICATE(CADR^.KIND>SEXPR);
            EQSYM  : APPLY:=EQQ(CADR, CADDR, ALIST);
            CARSYM: APPLY:=HEAD(CADR, ALIST);
            CDRSYM: APPLY:=TAIL(CADR, ALIST);
            CONSSYM: APPLY:=CONS(CADR, CADDR, ALIST);
            COPYSYM: APPLY:=COPY(CADR, ALIST);
            APPENSYM: APPLY:=APPEND(CADR, CADDR, ALIST);
            REPLASYM: APPLY:=REPLACH(CADR, CADDR, ALIST);
            REPLDSYM: APPLY:=REPLACT(CADR, CADDR, ALIST);
            NUMBPSYM: APPLY:=PREDICATE(CADR^.KIND=NUMERIC);
            ADDSYM  : APPLY:=MATH(CADR^.NVALUE+CADDR^.NVALUE);
            SUBSYM  : APPLY:=MATH(CADR^.NVALUE-CADDR^.NVALUE);
            TIMESYM : APPLY:=MATH(CADR^.NVALUE*CADDR^.NVALUE);
            DIVSYM  : APPLY:=MATH(CADR^.NVALUE DIV CADDR^.NVALUE);
            EVALSYM : APPLY:=EVAL(CADR,CADDR);
            APPLYSYM: BEGIN FN:=CADR; AGAIN:=TRUE;
                ALIST:=HEAD(TAIL(TAIL(ARGS)));
                ARGS:=CADDR  END;
          OTHERWISE ERROR(3)
          END END;
         PVALUE:
          IF FN=NILATOM THEN BEGIN AGAIN:=FALSE; APPLY:=NILATOM END
          ELSE ERROR(14);
         OTHERWISE ERROR(14)
       END
     END
   ELSE BEGIN PPTR:=GETPROP(HEAD(FN),[PFSUBR]);
     IF PPTR^.KIND=PFSUBR THEN
       BEGIN
       CADR:=HEAD(TAIL(FN)); CADDR:=HEAD(TAIL(TAIL(FN)));
       CASE PPTR^.RESSYM OF
         LABELSYM: BEGIN FN:=CADDR;
           ALIST:=CONS(CONS(CADR,CADDR),ALIST);
           END;
         FUNARSYM: BEGIN FN:=CADR; ALIST:=CADDR; END;
         LAMBDSYM: BEGIN AGAIN:=FALSE;
          APPLY :=EVAL(CADDR,BINDARGS(CADR,ARGS));
          END;
         OTHERWISE ERROR(12)
       END END
     ELSE FN:=EVAL(FN,ALIST);
     END;
   END;


  BEGIN (* EVAL *)
    CASE E^.KIND OF 
     NUMERIC : EVAL:=E;
     ANATOM  : BEGIN PPTR:=GETPROP(E,[PVALUE]);
        IF PPTR<>NOPROP THEN
          EVAL:=PPTR^.PNODE
        ELSE EVAL:=TAIL(LOOKUP(E,ALIST,TRUE));
      END;
     SEXPR   :
      BEGIN
       CAROFE:=E^.HEAD;
       IF CAROFE^.KIND=ANATOM THEN
        BEGIN PPTR:=GETPROP(CAROFE,[PFEXPR,PEXPR,PSUBR,PFSUBR]);
          CASE PPTR^.KIND OF
            PEXPR: EVAL:=APPLY(PPTR^.PNODE,EVLIS(E^.TAIL),ALIST);
            PSUBR: EVAL:=APPLY(CAROFE,EVLIS(E^.TAIL),ALIST);
            PFEXPR: EVAL:=APPLY(PPTR^.PNODE,
              CONS(TAIL(E),CONS(ALIST,NILATOM)),ALIST);
            PFSUBR:
              CASE PPTR^.RESSYM OF
                QUOTESYM: EVAL:=HEAD(E^.TAIL);
                CONDSYM : EVAL:=EVCON(E^.TAIL);
                FUNCTSYM: EVAL:=CONS(FNARGATOM,
                  CONS(HEAD(E^.TAIL),CONS(ALIST,NILATOM)));
                DEFINSYM: BEGIN E:=E^.TAIL; TEMP:=NILATOM;
                  WHILE E<>NILATOM DO
                    BEGIN CAROFE:=HEAD(E); E:=TAIL(E);
                      TEMP:=CONS(HEAD(CAROFE),TEMP);
                      SETPROP(PPTR,TEMP^.HEAD,PEXPR);
                     PPTR^.PNODE:=HEAD(TAIL(CAROFE));
                    END;
                  EVAL:=TEMP
                  END;
                SETQSYM: BEGIN
                  CAROFE:=HEAD(E^.TAIL);
                  CAAROFE:=LOOKUP(CAROFE,ALIST,FALSE);
                  TEMP:=EVAL(HEAD(TAIL(E^.TAIL)),ALIST);
                  IF CAAROFE<>NILATOM THEN
                    TEMP:=REPLACT(CAAROFE,TEMP)
                  ELSE BEGIN
                    SETPROP(PPTR,CAROFE,PVALUE);
                    PPTR^.PNODE:=TEMP
                    END;
                  EVAL:=CAROFE;
                  END;
              OTHERWISE ERROR(3)
              END;
            PNONE:
            EVAL:=EVAL(
              CONS(TAIL(LOOKUP(CAROFE,ALIST,TRUE)),E^.TAIL),ALIST);
          OTHERWISE ERROR(14)
          END;
        END
        ELSE EVAL:=APPLY(CAROFE,EVLIS(E^.TAIL),ALIST);
       END
     END
  END (* EVAL *);




  PROCEDURE INITIALIZE;

  VAR I:INTEGER; TEMP: SYMBEXPPTR;
    DEFWORD: RESERVDWORDS; PK: PROPKIND;

  BEGIN
    ALREADYPEEKED:=FALSE; READ(CH);
    NUMBEROFGCS:=0;
    FREENODES:=MAXNODE;
    FREELIST:=NIL;
    FOR I:=1 TO MAXNODE DO
     BEGIN NEW(NODELIST); NODELIST^.NEXT:=FREELIST;
       NODELIST^.HEAD:=FREELIST; NODELIST^.STATUS:=UNMARKED;
       FREELIST:=NODELIST
     END;

    NOPROP:=NIL; OBLIST:=NIL; NILATOM:=NIL;

    NILATOM:=FORMATOM('NIL       ');
    OBLIST^.TAIL^.HEAD:=NILATOM; OBLIST^.TAIL^.TAIL:=NILATOM;
    TATOM  :=FORMATOM('T         ');
    FINATOM:=FORMATOM('FIN       ');
    TEMP   :=FORMATOM('OBLIST    ');
    SETPROP(NOPROP,NILATOM,PNONE);
    SETPROP(PPTR,NILATOM,PVALUE); PPTR^.PNODE:=NILATOM;
    SETPROP(PPTR,TATOM,PVALUE); PPTR^.PNODE:=TATOM;
    SETPROP(OBPROP,TEMP,PVALUE);

    PK:=PFSUBR; I:=0;
    FOR DEFWORD:=LABELSYM TO PRED(LASTSYM) DO
     BEGIN TEMP:=FORMATOM(RESWORDS[DEFWORD]);
       IF DEFWORD=ATOMSYM THEN PK:=PSUBR;
       SETPROP(PPTR,TEMP,PK);
       PPTR^.RESSYM:=DEFWORD;
       IF DEFWORD=CONSSYM THEN I:=1;
       PPTR^.ARGS:=I
     END;
    FNARGATOM:=FORMATOM('FUNARG    ');

  END (* INITIALIZE *);


BEGIN (* LISP *)
  WRITELN('* EVAL * ');  INITIALIZE;  NEXTSYM; READEXPR(PTR);
  READLN; WRITELN;
  WHILE PTR<>FINATOM DO
    BEGIN
      ERRORS:=FALSE;
      WRITELN(' * value * ');
      OBPROP^.PNODE:=OBLIST;
      EVALVAL:=EVAL(PTR,NILATOM);
      IF NOT ERRORS THEN PRINTEXPR(EVALVAL);
      WRITELN; WRITELN;
      IF EOF THEN ERROR(11);
      PTR:=NIL;
      IF FREENODES < (MAXNODE DIV 3) THEN
        BEGIN GARBAGEMAN; WRITELN; WRITELN END;
      WRITELN(' * EVAL * '); NEXTSYM; READEXPR(PTR);
      READLN
    END;
  WRITELN; WRITELN;
  WRITELN(' Total number of garbage collections = ', NUMBEROFGCS:1,
          '.');
  WRITELN;
  WRITELN(' Free nodes left upon exit = ', FREENODES:1, '.');
  WRITELN
END (* LIST *).
